"""
 Copyright (c) 2022 Antonio Fernandez-Guerra

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to
 use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 the Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 """


import logging
from collapse_gtdb_tree.utils import (
    get_arguments,
)
import pandas as pd
from ete3 import Tree
import os
from pathlib import Path

log = logging.getLogger("my_logger")


def create_collapsed_tree_files(file, ranks):
    """
    Create collapsed tree files
    """
    # get file name from path
    file_name = Path(file).resolve().stem.split(".")[0]
    files = {}
    for rank in ranks:
        files[rank] = f"{file_name}_{rank}.tree"
    return files


def collapse_gtdb_tree_by_rank(tree, taxonomy, ranks, files, quoted_names=False):
    for rank in ranks:
        t = tree.copy()
        df = taxonomy[["genome_id", rank]]
        df = df.groupby([rank]).agg(lambda x: x.iloc[0]).reset_index()
        ids = df.set_index(
            "genome_id",
        ).to_dict()[rank]
        t.prune(list(ids.keys()))
        logging.info(f"Pruned tree for {rank} [leaves: {len(t.get_leaves()):,}]")
        # rename leaves
        for leaf in t.iter_leaves():
            leaf.name = ids[leaf.name]
        t.write(outfile=str(files[rank]), quoted_node_names=quoted_names, format=1)


def main():
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)s ::: %(asctime)s ::: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    args = get_arguments()
    logging.getLogger("my_logger").setLevel(
        logging.DEBUG if args.debug else logging.INFO
    )

    # Read taxonomy file
    logging.info(f"Reading taxonomy file {os.path.basename(args.taxonomy)}")
    taxonomy_file = pd.read_csv(
        args.taxonomy, sep="\t", names=["genome_id", "tax_string"]
    )
    ranks = ["domain", "phylum", "class", "order", "family", "genus", "species"]

    taxonomy_file[ranks] = taxonomy_file.tax_string.str.split(
        ";",
        expand=True,
    )
    logging.info(f"Read tree file {os.path.basename(args.tree)}")
    tree = Tree(args.tree, quoted_node_names=True, format=1)
    logging.info(f"Read tree with {len(tree.get_leaves()):,} leaves")
    tips = [x.name for x in tree.iter_leaves()]
    taxonomy_file = taxonomy_file[taxonomy_file["genome_id"].isin(tips)]

    files = create_collapsed_tree_files(args.tree, args.rank)
    # get list of ids by ranks
    collapse_gtdb_tree_by_rank(
        tree=tree,
        taxonomy=taxonomy_file,
        ranks=args.rank,
        files=files,
        quoted_names=args.quoted_names,
    )


if __name__ == "__main__":
    main()



import argparse
import sys
import gzip
import os
import logging
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull
from collapse_gtdb_tree import __version__
import time


log = logging.getLogger("my_logger")
log.setLevel(logging.INFO)
timestr = time.strftime("%Y%m%d-%H%M%S")


def is_debug():
    return logging.getLogger("my_logger").getEffectiveLevel() == logging.DEBUG


filters = ["breadth", "depth", "depth_evenness", "breadth_expected_ratio"]


# From https://stackoverflow.com/a/59617044/15704171
def convert_list_to_str(lst):
    n = len(lst)
    if not n:
        return ""
    if n == 1:
        return lst[0]
    return ", ".join(lst[:-1]) + f" or {lst[-1]}"


def get_compression_type(filename):
    """
    Attempts to guess the compression (if any) on a file using the first few bytes.
    http://stackoverflow.com/questions/13044562
    """
    magic_dict = {
        "gz": (b"\x1f", b"\x8b", b"\x08"),
        "bz2": (b"\x42", b"\x5a", b"\x68"),
        "zip": (b"\x50", b"\x4b", b"\x03", b"\x04"),
    }
    max_len = max(len(x) for x in magic_dict)

    unknown_file = open(filename, "rb")
    file_start = unknown_file.read(max_len)
    unknown_file.close()
    compression_type = "plain"
    for file_type, magic_bytes in magic_dict.items():
        if file_start.startswith(magic_bytes):
            compression_type = file_type
    if compression_type == "bz2":
        sys.exit("Error: cannot use bzip2 format - use gzip instead")
        sys.exit("Error: cannot use zip format - use gzip instead")
    return compression_type


def get_open_func(filename):
    if get_compression_type(filename) == "gz":
        return gzip.open
    else:  # plain text
        return open


# From: https://stackoverflow.com/a/11541450
def is_valid_file(parser, arg, var):
    if not os.path.exists(arg):
        parser.error("argument %s: The file %s does not exist!" % (var, arg))
    else:
        return arg


def get_ranks(parser, ranks, var):
    valid_ranks = ["all", "phylum", "class", "order", "family", "genus"]
    ranks = ranks.split(",")
    # check if ranks are valid
    for rank in ranks:
        if rank not in valid_ranks:
            parser.error(
                f"argument {var}: Invalid value {rank}.\Rank has to be one of {convert_list_to_str(valid_ranks)}"
            )
        if rank == "all":
            ranks = valid_ranks[1:]
    return ranks


defaults = {
    "rank": "all",
}

help_msg = {
    "tree": "GTDB tree file",
    "taxonomy": "GTDB taxonomy file",
    "rank": "Select taxonomic rank to collapse",
    "debug": "Print debug messages",
    "quoted_names": "Add quotes to the node and leaf names",
    "version": "Print program version",
}


def get_arguments(argv=None):
    parser = argparse.ArgumentParser(
        description="A simple tool to collapse a GTDB tree to a certain taxonomic rank",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-t",
        "--tree",
        type=lambda x: is_valid_file(parser, x, "--tree"),
        help=help_msg["tree"],
        required=True,
    )
    parser.add_argument(
        "-T",
        "--taxonomy",
        type=lambda x: is_valid_file(parser, x, "--taxonomy"),
        help=help_msg["taxonomy"],
        required=True,
    )
    # this is an argument to comma separated list of ranks for which to collapse
    parser.add_argument(
        "-r",
        "--rank",
        type=lambda x: get_ranks(parser, x, "--rank"),
        default=defaults["rank"],
        help=help_msg["rank"],
    )
    parser.add_argument(
        "--quoted-names",
        dest="quoted_names",
        action="store_true",
        help=help_msg["quoted_names"],
    )
    parser.add_argument(
        "--debug", dest="debug", action="store_true", help=help_msg["debug"]
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__,
        help=help_msg["version"],
    )
    args = parser.parse_args(None if sys.argv[1:] else ["-h"])
    return args


@contextmanager
def suppress_stdout():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)
